/*
 *	TTY_BSD -- terminal I/O routines for *BSD.
 *	Copyright (C) 1996-2000 Nils M Holm. All rights reserved.
 *	See the file LICENSE for conditions of use.
 *
 *	The TTYCTL class requires that these routines be
 *	placed in a shared object named 'ttyctl.so'.
 */

#include <stdio.h>
#include <stdlib.h>
#include <termcap.h>
#include <sgtty.h>
#include <unistd.h>
#include <sys/time.h>

#define K_HOME	327
#define K_WLFT	371
#define K_LEFT	331
#define K_RGHT	333
#define K_WRGT	372
#define K_END	335
#define K_BKSP	'\b'
#define K_DEL	339
#define K_KILL	127
#define K_INS	338
#define K_CR	'\r'
#define K_UP	328
#define K_DOWN	336
#define K_ESC	'\033'
#define K_PGUP	329
#define K_PGDN	337
#define K_CPGU	388
#define K_CPGD	374
#define K_F1	315
#define K_F2	316
#define K_F3	317
#define K_F4	318
#define K_F5	319
#define K_F6	320
#define K_F7	321
#define K_F8	322
#define K_F9	323
#define K_F10	324


#define CE	0
#define CL	1
#define CM	2
#define CS	3
#define SF	4
#define SR	5
#define SE	6
#define SO	7

static char	*tc[9][3] = {
	{	"ce",	NULL,	"erase line"			},
	{	"cl",	NULL,	"erase display"			},
	{	"cm",	NULL,	"cursor move"			},
	{	"cs",	NULL,	"change scrolling region"	},
	{	"sf",	NULL,	"scroll forward"		},
	{	"sr",	NULL,	"scroll reverse"		},
	{	"se",	NULL,	"end standout"			},
	{	"so",	NULL,	"standout"			},
	{	NULL	}
};

struct keycode {
	char	*name,
		*seq;
	int	code;
	char	*desc;
} kc[] = {
	{	"**",	"\033[B",	K_DOWN,	"vt100 down"	},
	{	"**",	"\033[D",	K_LEFT,	"vt100 left"	},
	{	"**",	"\033[C",	K_RGHT,	"vt100 right"	},
	{	"**",	"\033[A",	K_UP,	"vt100 up"	},
	{	"kd",	"\033[B",	K_DOWN,	"key down"	},
	{	"kl",	"\033[D",	K_LEFT,	"key left"	},
	{	"kr",	"\033[C",	K_RGHT,	"key right"	},
	{	"ku",	"\033[A",	K_UP,	"key up"	},
	{	"k1",	"\033Oq",	K_F1,	"key F1"	},
	{	"k2",	"\033Or",	K_F2,	"key F2"	},
	{	"k3",	"\033Os",	K_F3,	"key F3"	},
	{	"k4",	"\033Ot",	K_F4,	"key F4"	},
	{	"k5",	"\033Ou",	K_F5,	"key F5"	},
	{	"k6",	"\033Ov",	K_F6,	"key F6"	},
	{	"k7",	"\033Ow",	K_F7,	"key F7"	},
	{	"k8",	"\033Ox",	K_F8,	"key F8"	},
	{	"k9",	"\033Oy",	K_F9,	"key F9"	},
	{	"k0",	"\033Op",	K_F10,	"key F10"	},
	{	"kP",	"\033Or",	K_PGUP,	"key prev"	},
	{	"kN",	"\033Oq",	K_PGDN,	"key next"	},
	{	"kh",	"\033[H",	K_HOME,	"key home"	},
	{	"kH",	"\033[F",	K_END,	"key eol"	},
	{	"kI",	"\033On",	K_INS,	"key insert"	},
	{	"kD",	"\033D",	K_DEL,	"key delete"	},
	{	NULL	}
};

#define MAXLEVEL	64

static struct sgttyb	raw, *oldvec[MAXLEVEL];
static int		level = 0;

#define KBUFLEN	64
static char	key_buf[KBUFLEN], key_ptr, key_len;

int	tty_cols,
	tty_lines,
	tty_attr,
	tty_x,
	tty_y;

int	Ansicolor;


int tty_writec(c)
int	c;
{
	char	ch = c;

	write(0, &ch, 1);
	return(c);
}


void tty_writes(s)
char	*s;
{
	write(0, s, strlen(s));
}


void tty_clear() {
	tputs(tc[CL][1], 1, tty_writec);
}


void tty_clreol() {
	tputs(tc[CE][1], 1, tty_writec);
}


void tty_color(c)
int	c;
{
	int	ctab[] = { '0', '4', '2', '6', '1', '5', '3', '7' };
	int	f, b, h;
	char	buf[40];

	tty_attr = c;
	f = ctab[c & 7];
	b = ctab[(c>>4) & 7];
	if (Ansicolor) {
		h = ((c&8) == 8) + '0';
		sprintf(buf, "\033[%c;3%c;4%cm", (char) h, (char) f, (char) b);
		tputs(buf, 1, tty_writec);
	}
	else {
		if (b > f)
			tputs(tc[SO][1], 1, tty_writec);
		else
			tputs(tc[SE][1], 1, tty_writec);
	}
}


int tty_query() {
	struct fd_set	fdset;
	struct timeval	z;

	FD_ZERO(&fdset);
	FD_SET(0, &fdset);
	z.tv_sec = z.tv_usec = 0L;
	return(select(1, &fdset, NULL, NULL, &z) == 1);
}



int tty_readc() {
	char		ch;
	int		i;
	struct fd_set	fdset;
	struct timeval	z;

rescan:
	if (key_len > key_ptr) {
		return(key_buf[key_ptr++]);
	}

	read(1, &ch, 1);
	if (ch != K_ESC) return(ch);

	FD_ZERO(&fdset);
	FD_SET(0, &fdset);
	z.tv_sec = 0L;
	z.tv_usec = 300 * 1000L;
	if (select(1, &fdset, NULL, NULL, &z) != 1) return(ch);

	key_buf[0] = K_ESC; key_ptr = 1;
	while (1) {
		if (select(1, &fdset, NULL, NULL, &z) != 1) break;
		if (key_ptr >= KBUFLEN-1) break;
		read(1, &ch, 1);
		key_buf[key_ptr++] = ch;
		key_buf[key_ptr] = 0;
		for (i=0; kc[i].name; i++) {
			if (kc[i].seq && !strcmp(key_buf, kc[i].seq)) {
				key_len = key_ptr = 0;
				return(kc[i].code);
			}
		}
	}

	key_len = key_ptr; key_ptr = 0;
	goto rescan;
}


void tty_move(x, y)
int	x, y;
{
	tty_x = x;
	tty_y = y;
	tputs(tgoto(tc[CM][1], tty_x, tty_y), 1, tty_writec);
}


void tty_scroll(top, bot)
int	top, bot;
{
	tputs(tparm(tc[CS][1], top, bot-1), bot-top, tty_writec);
	tty_move(0, bot-1);
	tputs(tc[SF][1], bot-top, tty_writec);
	tputs(tparm(tc[CS][1], 0, tty_lines-1), tty_lines, tty_writec);
}


void tty_rscroll(top, bot)
int	top, bot;
{
	tputs(tparm(tc[CS][1], top, bot-1), bot-top, tty_writec);
	tty_move(0, top);
	tputs(tc[SR][1], bot-top, tty_writec);
	tputs(tparm(tc[CS][1], 0, tty_lines-1), tty_lines, tty_writec);
}


void tty_init() {
	char		*tname, *tp, tbuf[1024], *t;
	static char	tsbuf[1024];
	int		i;

	if (level >= MAXLEVEL-1) {
		fprintf(stderr, "TTYCTL: too many nested init()s\n");
		exit(EXIT_FAILURE);
	}
	if ((oldvec[level++] = (struct sgttyb *) malloc(sizeof(struct sgttyb)))
	== NULL) {
		fprintf(stderr, "TTYCTL: out of memory\n");
		exit(EXIT_FAILURE);
	}
	if ((tname = getenv("TERM")) == NULL) {
		fprintf(stderr, "TTYCTL: the TERM variable is not set\n");
		exit(EXIT_FAILURE);
	}

	t = getenv("ANSICOLOR");
	if (t != NULL) {
		Ansicolor = !strcmp(t, "no")? 0:
			!strcmp(t, "No")? 0:
			!strcmp(t, "NO")? 0: 1;
	}
	else {
		Ansicolor = strstr(tname, "color") != NULL;
	}

	switch(tgetent(tbuf, tname)) {
		case -1: fprintf(stderr,
			 "TTYCTL: could not open termcap database");
			 exit(EXIT_FAILURE);
		case 0:  fprintf(stderr,
			"TTYCTL: unknown terminal type: %s\n", tname);
			 exit(EXIT_FAILURE);
	}

	if (	(tty_cols = tgetnum("co")) == -1 ||
		(tty_lines = tgetnum("li")) == -1
	) {
		fprintf(stderr, "TTYCTL: cannot determine screen size\n");
		exit(EXIT_FAILURE);
	}

	if (tty_lines > 25) tty_lines = 25;	/*XXX*/

	tp = tsbuf;
	for (i=0; tc[i][0]; i++) {
		if ((tc[i][1] = tgetstr(tc[i][0], &tp)) == NULL) {
			fprintf(stderr,
			"TTYCTL: your terminal is lacking the"
			" '%s' capability\n", tc[i][2]);
			exit(EXIT_FAILURE);
		}
	}

	for (i=0; kc[i].name; i++) {
		if ((t = tgetstr(kc[i].name, &tp)) != NULL)
			kc[i].seq = t;
	}

	tty_x = 0;
	tty_y = 0;
	tty_attr = 0x07;

	key_len = key_ptr = 0;

	gtty(0, oldvec[level-1]);
	gtty(0, &raw);
	raw.sg_flags |= RAW;
	raw.sg_flags &= ~(ECHO|CRMOD);
}


void tty_mode(rawmode) {
	stty(0, rawmode? &raw: oldvec[level-1]);
}


void tty_fini() {
	if (level < 1) {
		fprintf(stderr, "TTYCTL: redundant fini()\n");
		exit(EXIT_FAILURE);
	}
	tty_mode(0);
	free(oldvec[--level]);
}


int tty_nlines() { return tty_lines; }

int tty_ncols() { return tty_cols; }

int tty_cancolor() { return Ansicolor; }

